from __future__ import annotations  # PEP 585

import multiprocessing
import re

from .zettel import Zettel
from .zettelkasten import Zettelkasten


def find_broken_links(
    zk: Zettelkasten, nth=4
) -> list[tuple[Zettel, list[tuple[str, str]]]]:
    with multiprocessing.Pool(nth) as pool:
        rs = pool.map(func, zk.zettels)
    rez = []
    for r in rs:
        if not r:
            continue
        z, tus = r
        tus = [tu for tu in tus if tu[0] not in zk]
        if tus:
            rez.append((z, tus))
    return rez


def func(z: Zettel) -> tuple[Zettel, list[tuple[str, str]]] | None:
    recs = re.findall(r"\[\[id:(.+?)\]\[(.+?)\]\]", z.orgnode().dumps())
    # recs = [(uuid, title) for uuid, title in recs if uuid not in zk]
    return (z, recs) if recs else None





from __future__ import annotations  # PEP 585

import os.path
import configparser
from typing import Mapping

OPE = os.path.expanduser

CONFIG_FILE_LOCATIONS = [
    OPE("~/.orgee-roam.ini"),
    OPE("~/.config/orgee-roam.ini"),
]
DEFAULT_DICT: dict = {
    "roam_cache": "~/.orgee-roam-cache.json",
    "zettelkasten_root": "~/orgee_zettelkasten",
}
PATH_KEYS = ["roam_cache", "zettelkasten_root"]


def get_config(fn: str | None = None) -> tuple[dict[str, str], str | None]:
    locations = [fn] if fn else CONFIG_FILE_LOCATIONS
    for loc in locations:
        if dic := parse_file(loc):
            return (dic, loc)
    return (update_defaults(), None)


def parse_file(fn: str) -> dict[str, str] | None:
    if not os.path.isfile(fn):
        return None
    config = configparser.ConfigParser()
    config.read(fn)
    return update_defaults(config.defaults())


def update_defaults(dic: Mapping[str, str] | None = None) -> dict[str, str]:
    rez = DEFAULT_DICT.copy()
    if dic:
        rez.update(dic)
    for k in PATH_KEYS:
        rez[k] = OPE(rez[k])
    return rez



from __future__ import annotations  # PEP 585

import logging

from .zettelkasten import Zettelkasten
from .zettel import Zettel
from .create_zettel import create_zettel
from .zettel_org_heading import make_zettel_org_heading


def make_list_zettel(
    zettels: list[Zettel],
    title: str,
    fn: str,
    zk: Zettelkasten | None = None,
    overwrite=True,
):
    if not zk:
        zk = Zettelkasten()
    root_zettel = create_zettel(
        title=f"{title} ({len(zettels)} zettels)",
        properties=[("ROAM_EXCLUDE", "t")],
        body=[""],
        filename=fn,
        overwrite=overwrite,
    )
    root = root_zettel.orgnode()
    for zettel in zettels:
        node = make_zettel_org_heading(zettel)
        root.add_child(node)
    root.dump(root_zettel.filename)
    logging.info("Saved %d links to %s", len(zettels), root_zettel.filename)



from __future__ import annotations  # PEP 585

import datetime
import logging
import os.path
import re
import time
from dataclasses import dataclass, field

from orgee.parse_org import parse_org_file
from orgee.orgnode import OrgNode
from orgee.util import prop_by_key, first_prop_by_key, clean_text


@dataclass
class Zettel:
    uuid: str
    title: str
    filename: str
    updated_ts: float
    lastchecked_ts: float
    # Hash of the underlying OrgNode
    zettel_hash: str
    level: int
    lineno: int
    tags: set[str] = field(default_factory=set)
    all_tags: set[str] = field(default_factory=set)
    aliases: set[str] = field(default_factory=set)
    olp: list[str] = field(default_factory=list)
    properties: list[tuple[str, str]] = field(default_factory=list)

    @staticmethod
    def from_rec(rec: dict) -> Zettel:
        return Zettel(
            uuid=rec["uuid"],
            title=rec["title"],
            filename=rec["filename"],
            updated_ts=rec["updated_ts"],
            lastchecked_ts=rec["lastchecked_ts"],
            zettel_hash=rec["zettel_hash"],
            level=rec["level"],
            lineno=rec["lineno"],
            tags=set(rec["tags"]),
            all_tags=set(rec["all_tags"]),
            aliases=set(rec["aliases"]),
            olp=rec["olp"],
            properties=rec["properties"],
        )

    def to_rec(self) -> dict:
        rec = {
            "uuid": self.uuid,
            "title": self.title,
            "filename": self.filename,
            "updated_ts": self.updated_ts,
            "updated_dt": datetime.datetime.fromtimestamp(
                self.updated_ts
            ).isoformat(),
            "lastchecked_ts": self.lastchecked_ts,
            "zettel_hash": self.zettel_hash,
            "level": self.level,
            "lineno": self.lineno,
            "tags": list(self.tags),
            "all_tags": list(self.all_tags),
            "aliases": list(self.aliases),
            "olp": self.olp,
            "properties": self.properties,
        }
        return rec

    @staticmethod
    def from_org_file(fn: str) -> list[Zettel]:
        root = parse_org_file(fn)
        zettels: list = []
        updated_ts = os.path.getmtime(fn)
        for node in root.recurse_nodes():
            if node.prop_by_key("ID"):
                zettel = Zettel.from_orgnode(
                    node=node, filename=fn, updated_ts=updated_ts
                )
                if zettel:
                    zettels.append(zettel)
        return zettels

    @staticmethod
    def root_from_org_file(fn: str) -> Zettel | None:
        root = parse_org_file(fn)
        return Zettel.from_orgnode(
            root, filename=fn, updated_ts=os.path.getmtime(fn)
        )

    @staticmethod
    def from_orgnode(
        node: OrgNode, filename: str, updated_ts: float | None = None
    ) -> Zettel | None:
        xc = node.prop_by_key("ROAM_EXCLUDE", parse=True)
        if xc and xc[0] is True:
            return None

        uuid = node.prop_by_key("ID")[0]
        if not updated_ts:
            updated_ts = time.time()
        zettel_hash = node.node_hash()
        return Zettel(
            uuid=uuid,
            title=node.title,
            filename=filename,
            updated_ts=updated_ts,
            lastchecked_ts=updated_ts,
            zettel_hash=zettel_hash,
            level=node.actual_level(),
            lineno=node.lineno,
            tags=node.tags,
            all_tags=node.all_tags(),
            aliases=set(node.prop_by_key("ROAM_ALIASES", parse=True)),
            olp=node.olp(),
            properties=node.properties,
        )

    def olp_str(self) -> str:
        return " → ".join(map(clean_text, self.olp))

    def orgnode(self, root: OrgNode | None = None) -> OrgNode:
        if not root:
            root = parse_org_file(self.filename)
        assert root
        if self.level == 0:
            return root
        else:
            node = root.find_olp(self.olp[1:])
            if not node:
                raise Exception(
                    f"Could not find node {self.olp} in {self.filename}"
                )
            return node

    def prop_by_key(self, key: str, parse=True, case_insensitive=True) -> list:
        return prop_by_key(
            key=key,
            props=self.properties,
            parse=parse,
            case_insensitive=case_insensitive,
        )

    def first_prop_by_key(
        self, key: str, parse=True, case_insensitive=True
    ) -> str | None:
        return first_prop_by_key(
            key=key,
            props=self.properties,
            parse=parse,
            case_insensitive=case_insensitive,
        )

    def creation_ts(self, self_correct=True, verbose=True) -> float:
        def correct(ts: float):
            node.properties.append(("CREATED_TS", str(int(ts))))
            if rm := node.root_meta:
                if not rm.first_other_meta_by_key("CREATED"):
                    dt = datetime.datetime.fromtimestamp(ts)
                    rm.other_meta.append(
                        (
                            "CREATED",
                            "%s %s"
                            % (
                                dt.strftime("%a %d %b %Y %H:%M:%S"),
                                dt.astimezone().tzinfo,
                            ),
                        )
                    )
            node.dump_root(self.filename)
            # Update zettel properties
            self.properties = node.properties
            if verbose:
                logging.info("Added CREATED_TS prop to %s", self.olp_str())

        # Try the CREATED_TS property
        if s := self.first_prop_by_key("CREATED_TS"):
            return float(s)
        # Try the created fileprop
        node = self.orgnode()
        if rm := node.root_meta:
            if s := rm.first_other_meta_by_key("created", parse=False):
                if m := re.match(r".+\((\d+)\)", s):
                    ts = float(m.groups()[0])
                    if self_correct:
                        correct(ts)
                    return ts
        logging.warning("Zettel [%s] has no CREATED_TS!", self.olp_str())
        return 0



from __future__ import annotations  # PEP 585

import re

from .zettel import Zettel
from .zettelkasten import Zettelkasten


def references_dict(zk: Zettelkasten) -> dict[str, list[Zettel]]:
    dic: dict = {}
    for zettel in zk.zettels:
        s = zettel.orgnode().dumps()
        uuids = re.findall(r"\[\[id:(.+?)\]\[.+?\]\]", s)
        if uuids:
            for uuid in uuids:
                dic.setdefault(uuid, []).append(zettel)
    return dic


def references_dict2(zk: Zettelkasten) -> dict[str, list[Zettel]]:
    dic: dict = {}
    for zettel in zk.zettels:
        s = zettel.orgnode().dumps()
        uuids = re.findall(r"\[\[id:(.+?)\]\[.+?\]\]", s)
        if uuids:
            for uuid in uuids:
                dic.setdefault(uuid, []).append(zettel)
    return dic



from __future__ import annotations

import datetime
import logging

from orgee.orgnode import OrgNode

from .create_zettel import create_zettel
from .zettel import Zettel
from .zettelkasten import Zettelkasten


def split_zettel_container(zid: str, zk: Zettelkasten) -> list[Zettel]:
    def parse_child(node: OrgNode) -> Zettel | None:
        if "ignore" in node.tags:
            return None
        zid = node.first_prop_by_key("ID")
        if not zid:
            logging.error("Node %s has no ID!", node.olp_str())
        assert zid
        cts = node.first_prop_by_key("CREATED_TS")
        assert cts
        dt = datetime.datetime.fromtimestamp(float(cts))
        aliases = set(node.prop_by_key("ROAM_ALIASES", parse=True))
        if tags := node.all_tags():
            file_other_meta = [("FILETAGS", " ".join(tags))]
        else:
            file_other_meta = []
        zettel = create_zettel(
            title=node.title,
            aliases=aliases,
            properties=node.properties,
            body=node.body,
            children=node.children,
            file_other_meta=file_other_meta,
            dt=dt,
            zid=zid,
        )
        return zettel

    zettel = zk[zid]
    root = zettel.orgnode()
    zs = list(filter(None, map(parse_child, root.children)))
    root.children = []
    root.dump_root(zettel.filename)
    logging.info(
        "Split container into %d zettel%s", len(zs), "s" if len(zs) > 1 else ""
    )
    return zs



from __future__ import annotations  # PEP 585

import json
import logging
import os.path
from collections.abc import MutableMapping
from typing import Iterator, ValuesView

from orgee.util import is_org_file

from .config import get_config
from .zettel import Zettel

VERBOSE_LIMIT = 100


class Zettelkasten(MutableMapping):
    def __init__(
        self,
        roam_cache: str | None = None,
        zettelkasten_root: str | None = None,
        auto_update=False,
    ) -> None:
        config, _ = get_config()
        self.roam_cache = roam_cache if roam_cache else config["roam_cache"]
        self.zettelkasten_root = (
            zettelkasten_root
            if zettelkasten_root
            else config["zettelkasten_root"]
        )
        self._dic: dict[str, Zettel] | None = None
        self._dics_by_prop: dict[str, dict[str, Zettel]] = {}
        if auto_update:
            self.update_cache()

    def __getitem__(self, k: str) -> Zettel:
        return self.dic[k]

    def __setitem__(self, k: str, v: Zettel) -> None:
        self.dic[k] = v

    def __delitem__(self, k: str) -> None:
        del self.dic[k]

    def __iter__(self):
        return iter(self.dic)

    def __len__(self) -> int:
        return len(self.dic)

    @property
    def dic(self) -> dict[str, Zettel]:
        """
        All the zettels in the zettelkasten indexed by id
        """
        if self._dic is None:
            self._dic = self.load_json()
        return self._dic

    @property
    def zettels(self) -> ValuesView[Zettel]:
        return self.dic.values()

    def zettel_filenames(self) -> set[str]:
        return {zettel.filename for zettel in self.zettels}

    def file_zettels(self) -> dict[str, list[Zettel]]:
        dic: dict = {}
        for zettel in self.zettels:
            dic.setdefault(zettel.filename, []).append(zettel)
        return dic

    def update_cache(self) -> int:
        all_files = set(self.org_files())
        existing_files = self.zettel_filenames()
        deleted_files = existing_files - all_files
        fndic = self.file_zettels()
        changes = 0

        if deleted_files:
            be_quiet = len(deleted_files) > VERBOSE_LIMIT
            # print(f"Be quiet deleted={be_quiet}")
            if be_quiet:
                logging.info(
                    "Deleting %d file%s",
                    len(deleted_files),
                    "s" if len(deleted_files) > 1 else "",
                )
            for fn in deleted_files:
                for zettel in fndic[fn]:
                    if not be_quiet:
                        logging.info(
                            "Removing «%s»", self.dic[zettel.uuid].olp_str()
                        )
                    del self.dic[zettel.uuid]
                    changes += 1
                del fndic[fn]

        fns = [
            fn
            for fn, zettels in fndic.items()
            if os.path.getmtime(fn)
            > max(zettel.lastchecked_ts for zettel in zettels)
        ]
        if fns:
            be_quiet = len(fns) > VERBOSE_LIMIT
            # print(f"Be quiet changed={be_quiet}")
            if be_quiet:
                logging.info(
                    "Updating %d file%s", len(fns), "s" if len(fns) > 1 else ""
                )
            for fn in fns:
                uts = os.path.getmtime(fn)
                zettels = fndic[fn]
                zettels2 = Zettel.from_org_file(fn)
                uuids = {zettel.uuid for zettel in zettels}
                uuids2 = {zettel.uuid for zettel in zettels2}
                for uuid in uuids - uuids2:
                    if not be_quiet:
                        logging.info("Removing %s", self.dic[uuid].olp_str())
                    del self.dic[uuid]
                    changes += 1
                for zettel in zettels2:
                    uuid = zettel.uuid
                    if zettel0 := self.dic.get(uuid):
                        if zettel.zettel_hash != zettel0.zettel_hash:
                            olp0, olp = zettel0.olp_str(), zettel.olp_str()
                            if olp != olp0:
                                if not be_quiet:
                                    logging.info(
                                        "Updated «%s» → «%s»", olp0, olp
                                    )
                            else:
                                if not be_quiet:
                                    logging.info("Updated «%s»", olp)
                        else:
                            zettel.updated_ts = zettel0.updated_ts
                    else:
                        if not be_quiet:
                            logging.info("Adding «%s»", zettel.olp_str())

                    zettel.lastchecked_ts = uts
                    self.dic[uuid] = zettel
                    changes += 1

        new_files = all_files - existing_files
        if new_files:
            be_quiet = len(new_files) > VERBOSE_LIMIT
            # print(f"Be quiet new={be_quiet}")
            if be_quiet:
                logging.info(
                    "Adding %d new file%s",
                    len(new_files),
                    "s" if len(new_files) > 1 else "",
                )
            for fn in new_files:
                zettels = Zettel.from_org_file(fn)
                for zettel in zettels:
                    uuid = zettel.uuid
                    if uuid in self.dic:
                        print(uuid)
                        print(zettel)
                        print(self.dic[uuid])
                        raise Exception("Duplicate ID!")
                    self.dic[uuid] = zettel
                    changes += 1
                    if not be_quiet:
                        logging.info("Adding «%s»", zettel.olp_str())

        if changes:
            logging.info(
                "%d node%s changed", changes, "s" if changes > 1 else ""
            )
            self.save_json()
        else:
            logging.info("No node changed")
        return changes

    def org_files(self) -> Iterator[str]:
        """
        Return a list of all org files in directory
        """
        for root, _, files in os.walk(self.zettelkasten_root):
            for fn in files:
                if not is_org_file(fn):
                    continue
                yield os.path.join(root, fn)

    def save_json(self):
        recs = [
            zettel.to_rec()
            for zettel in sorted(
                self.zettels, key=lambda n: n.updated_ts, reverse=True
            )
        ]
        with open(self.roam_cache, "w", encoding="utf-8") as fh:
            json.dump(recs, fh, indent=2, ensure_ascii=False)

    def load_json(self) -> dict[str, Zettel]:
        if os.path.isfile(self.roam_cache):
            with open(self.roam_cache, "r", encoding="utf-8") as fh:
                recs = json.load(fh)
                zettels = map(Zettel.from_rec, recs)
                return {zettel.uuid: zettel for zettel in zettels}
        else:
            return {}

    def dict_by_prop(
        self, key: str, check_unique=True, use_memoized=True
    ) -> dict[str, Zettel]:
        """
        Return the zettelkasten indexed by a node property
        """
        if use_memoized and (dic := self._dics_by_prop.get(key)):
            return dic
        # pairs = [
        #     (prop, zettel)
        #     for zettel in self.zettels
        #     if (prop := zettel.first_prop_by_key(key))
        # ]
        # Allow a zettel to have multiple props
        pairs = []
        for zettel in self.zettels:
            props = zettel.prop_by_key(key)
            for prop in props:
                pairs.append((prop, zettel))

        dic = dict(pairs)
        if len(dic) == len(pairs) or not check_unique:
            self._dics_by_prop[key] = dic
            return dic
        else:
            d: dict = {}
            for ca, z in pairs:
                d.setdefault(ca, []).append(z)
            for ca, zs in d.items():
                if len(zs) > 1:
                    print(f"{ca}: {', '.join(z.olp_str() for z in zs)}")
            raise Exception("Some props are in multiple zettels!")



from __future__ import annotations

# import os.path

# from orgtools.util import clean_text
from .zettel import Zettel
from .create_zettel import create_zettel
from .zettel_org_heading import make_zettel_org_heading


class ZettelFinder:
    def __init__(self, zettels: list[Zettel]):
        self.zettels = zettels

    def filter(self, filter_func) -> ZettelFinder:
        return ZettelFinder(zettels=list(filter(filter_func, self.zettels)))

    def make_finder_file(self, fn: str, title: str, add_info_func=None):
        root_zettel = create_zettel(
            title=f"{title} ({len(self.zettels)})",
            properties=[("ROAM_EXCLUDE", "t")],
            body=[""],
            filename=fn,
            overwrite=True,
        )
        root = root_zettel.orgnode()
        for zettel in self.zettels:
            root.add_child(
                make_zettel_org_heading(
                    zettel,
                    use_id=True,
                    add_info_func=add_info_func,
                )
            )
        root.dump_root(root_zettel.filename)

    # def make_links_file(self, fn: str, title: str = None):
    #     root_zettel = create_zettel(
    #         title=title if title else f"Zettel Links ({len(self.zettels)} zettels)",
    #         body=[""],
    #         filename=fn,
    #         overwrite=True,
    #         root=os.path.expanduser("~/.emacs.d"),
    #     )
    #     body: list = []
    #     for zettel in self.zettels:
    #         body.append("")
    #         h = f"{zettel.olp_str()} #zz"
    #         if zettel.tags:
    #             h += " " + " ".join(f"#{tag}" for tag in sorted(zettel.tags))
    #         body.append(h)
    #         names = [zettel.title]
    #         if zettel.aliases:
    #             names.extend(zettel.aliases)
    #         url = f"id:{zettel.uuid}"
    #         for name in names:
    #             body.append(f"[[{url}][{clean_text(name)}]]")
    #     root = root_zettel.orgnode()
    #     root.body = body
    #     root.dump_root(root_zettel.filename)



from __future__ import annotations  # PEP 585

import datetime
import logging
import random
import re
from urllib.parse import unquote

import wikipediaapi  # type:ignore

from orgee.orgnode import OrgNode

from .zettelkasten import Zettelkasten
from .zettel import Zettel

STALE_ND = 180
PROB_UPDATE = 0.1


def update_wikipedia_org_subtrees(
    zettel: Zettel,
    wps: dict | None = None,
    force: bool = False,
    stale_nd: int = STALE_ND,
    prob_update: float = PROB_UPDATE,
) -> bool:
    name = zettel.olp_str()
    if not wps:
        wps = wikiprops(zettel)
    if not wps:
        logging.warning("No wikipedia property in %s", name)
        return False
    node = zettel.orgnode()
    changed = False
    filename = zettel.filename

    # Update page for each lang
    for lang, tu in wps.items():
        title, update_prop_p = tu
        pk = wikipedia_prop_key(lang)

        # Automatically find page from other languages
        if not title:
            for lang2, tu2 in wps.items():
                if title2 := tu2[0]:
                    wi = wikipediaapi.Wikipedia(lang2)
                    page = wi.page(title2)
                    if not (rec := page.langlinks.get(lang)):
                        logging.error(
                            "No page in %s for %s(%s)", lang, title2, lang2
                        )
                        continue
                    title = rec.title
                    update_prop(node=node, pk=pk, pv=title)
                    node.dump_root(filename)
                    changed = True
                    logging.info("Updated wiki title for %s to %s", lang, title)
                    break
            if not title:
                logging.error(
                    "Could not find link in %s from another language for «%s»",
                    lang,
                    zettel.olp_str(),
                )
                continue

        if update_prop_p:
            update_prop(node=node, pk=pk, pv=title)
            logging.info("Updated wiki title for %s to %s", lang, title)
            node.dump_root(filename)
            changed = True

        url = wiki_url(title=title, lang=lang)
        wnode = node.find_child_by_title(url=url)
        # Don't update existing subtree unless forced
        if wnode:
            if not force:
                ret_node = wnode.find_child_by_title(
                    title="Retrieved", startswith=True
                )
                if ret_node:
                    s = ret_node.title
                    m = re.match(r"Retrieved\s*\[(\d{4}-\d{2}-\d{2}).*\]", s)
                    if not m:
                        raise Exception(f"Could not parse {s} at {name}")
                    dt = datetime.datetime.strptime(
                        m.groups()[0], "%Y-%m-%d"
                    ).date()
                    if (datetime.date.today() - dt).days < stale_nd:
                        continue
                    elif random.random() > prob_update:
                        continue

        nwnode, actual_title = mk_wnode(lang=lang, title=title)
        nwnode.level = node.actual_level() + 1
        if not wnode and actual_title != title:
            title = actual_title
            update_prop(node=node, pk=pk, pv=title)
            node.dump_root(filename)
            changed = True
            url = wiki_url(title=title, lang=lang)
            wnode = node.find_child_by_title(url=url)
        if not wnode:
            node.add_child(nwnode)
        else:
            if subtrees_equal_p(nwnode, wnode):
                continue
            idx = node.find_child_by_title_index
            node.add_child(node=nwnode, position=idx, replace=True)
        node.tags.add("wikipedia")
        node.dump_root(filename)
        logging.info("Updated subtree for %s(%s) in node %s", title, lang, name)
        changed = True

    return changed


def update_prop(node: OrgNode, pk: str, pv: str):
    node.properties = [(k, v) for k, v in node.properties if k != pk]
    node.properties.append((pk, pv))
    logging.info("Updating property %s to %s", pk, pv)


def subtrees_equal_p(node1: OrgNode, node2: OrgNode) -> bool:
    b1 = ("\n".join(node1.body)).strip()
    b2 = ("\n".join(node2.body)).strip()
    return b1 == b2


def mk_wnode(lang: str, title: str) -> tuple[OrgNode, str]:
    wi = wikipediaapi.Wikipedia(lang)
    page = wi.page(title)
    n = OrgNode()
    try:
        url = unquote(page.fullurl)
    except:
        print(page)
        raise
    n.title = "[[%s][Wikipedia(%s): %s]]" % (url, lang, page.title)
    n.tags.add("auto")
    n.body.append(page.summary)
    dt = datetime.date.today().strftime("%Y-%m-%d %a")
    n.add_child(OrgNode(title=f"Retrieved [{dt}]"))
    m = re.match(r"https://.+/wiki/(.+)", url)
    assert m
    return n, m.groups()[0]


def wikiprops(zettel: Zettel) -> dict[str, tuple[str, bool]]:
    def match(v):
        if m := re.match(r"https://.+/wiki/(.+)", v):
            return unquote(m.groups()[0]).strip()
        else:
            return None

    dic = {}
    for k, v in zettel.properties:
        if k.startswith("WIKIPEDIA_PAGE"):
            if k.startswith("WIKIPEDIA_PAGE_"):
                lang = k[-2:].lower()
            elif k == "WIKIPEDIA_PAGE":
                lang = "en"
            else:
                raise Exception(f"Unexpected property {k}")

            v = v.strip('"')
            if title := match(v):
                dic[lang] = (title, True)
            else:
                dic[lang] = (v, False)
    return dic  # type:ignore


def wikinodes(zk: Zettelkasten) -> list[tuple[Zettel, dict]]:
    return [
        (zettel, wps) for zettel in zk.zettels if (wps := wikiprops(zettel))
    ]


def wikipedia_prop_key(lang: str | None = None) -> str:
    if not lang or lang == "en":
        k = "WIKIPEDIA_PAGE"
    else:
        k = f"WIKIPEDIA_PAGE_{lang.upper()}"
    return k


def wiki_url(title: str, lang: str | None = None):
    if not lang:
        lang = "en"
    return f"https://{lang}.wikipedia.org/wiki/{title}"



from __future__ import annotations

import datetime

from .zettelkasten import Zettelkasten
from .finder import ZettelFinder
from .zettel import Zettel


def make_finder_files():
    # def add_info_func(z: Zettel) -> str:
    #     ts = z.updated_ts
    #     if ts:
    #         return datetime.datetime.fromtimestamp(ts).strftime(
    #             "(%a %d %b %Y, %H:%M:%S)"
    #         )
    #     else:
    #         return ""
    def add_info_func(z: Zettel) -> str:
        uts = z.updated_ts
        cts = z.creation_ts()
        return "(%s | /%s/)" % (
            datetime.datetime.fromtimestamp(uts).strftime("%a %d %b %Y, %H:%M")
            if uts
            else "–",
            datetime.datetime.fromtimestamp(cts).strftime("%d %b %y") if cts else "–",
        )

    zk = Zettelkasten(auto_update=True)
    zettels = sorted(zk.zettels, key=lambda z: z.updated_ts, reverse=True)
    zf = ZettelFinder(zettels=zettels)
    zf.make_finder_file(
        fn="zettel-finder.org",
        title="Nodes by updated timestamp",
        add_info_func=add_info_func,
    )
    # zf.make_links_file(fn="zettel-finder.org")
    zettels = restrict_zettels(zettels)
    zf = ZettelFinder(zettels=zettels)
    zf.make_finder_file(
        fn="zettel-finder-restricted.org",
        title="Restricted nodes by updated timestamp",
        add_info_func=add_info_func,
    )
    # zf.make_links_file(fn="zettel-finder-restricted.org")


def make_finder_by_creation_ts():
    def add_info_func(z: Zettel) -> str:
        ts = z.creation_ts()
        if ts:
            return datetime.datetime.fromtimestamp(ts).strftime("(%a %d %b %Y, %H:%M)")
        else:
            return ""

    zk = Zettelkasten(auto_update=True)
    zettels = sorted(zk.zettels, key=lambda z: z.creation_ts(), reverse=True)
    zf = ZettelFinder(zettels=zettels)
    zf.make_finder_file(
        fn="zettel-finder-by-ts.org",
        title="Nodes by creation timestamp",
        add_info_func=add_info_func,
    )
    zettels = restrict_zettels(zettels)
    zf = ZettelFinder(zettels=zettels)
    zf.make_finder_file(
        fn="zettel-finder-by-ts-restricted.org",
        title="Restricted nodes by creation timestamp",
        add_info_func=add_info_func,
    )


def restrict_zettels(zettels: list[Zettel]) -> list[Zettel]:
    exclude_tags = {
        "album",
        "article",
        "band",
        "book",
        "character",
        "country",
        "movie",
        "painting",
        "paper",
        "person",
        "song",
        "stock",
        "video",
        "webclip",
        "youtube",
    }
    return [z for z in zettels if not exclude_tags & z.all_tags]



from __future__ import annotations  # PEP 585

import datetime
import logging
import os.path
import time
import unicodedata
import uuid

from slugify import slugify  # type:ignore

from orgee.orgnode import OrgNode
from orgee.util import dump_property

from .config import get_config
from .zettel import Zettel

FORBIDDEN_CHARS = '/\\:*"?<>|'


def create_zettel(
    title: str,
    aliases: set[str] | None = None,
    tags: set[str] | None = None,
    properties: list[tuple[str, str]] | None = None,
    body: list[str] | None = None,
    children: list[OrgNode] | None = None,
    parent: Zettel | None = None,
    file_properties: list[str] | None = None,
    file_other_meta: list[tuple[str, str]] | None = None,
    dt: datetime.datetime | None = None,
    root: str | None = None,
    filename: str | None = None,
    zid: str | None = None,
    overwrite=False,
) -> Zettel:
    if not dt:
        dt = datetime.datetime.now()
    if not root:
        config, _ = get_config()
        root = config["zettelkasten_root"]
    if not zid:
        zid = str(uuid.uuid4())

    node = OrgNode(title=title, is_root=not bool(parent))
    if properties:
        node.properties = properties
    node.replace_prop("ID", zid)
    node.replace_prop("CREATED_TS", str(int(dt.timestamp())))
    # node.properties.append(("ID", zid))
    # node.properties.append(("CREATED_TS", str(int(dt.timestamp()))))
    if aliases:
        node.replace_prop("ROAM_ALIASES", dump_property(list(aliases)))
    if tags:
        node.tags = tags
    if body:
        node.body = body
    if children:
        for child in children:
            node.add_child(child)

    if parent:
        if filename:
            raise Exception("Child node cannot set the parent filename")
        filename = parent.filename
        parent_node = parent.orgnode()
        parent_node.add_child(node)
        parent_node.dump_root(filename)
        logging.info("Added node %s to node %s", title, parent.olp_str())
    else:
        rm = node.root_meta
        assert rm
        rm.other_meta.append(
            (
                "CREATED",
                "%s %s"
                % (dt.strftime("%a %d %b %Y %H:%M"), dt.astimezone().tzinfo),
            )
        )
        if file_properties:
            rm.properties = file_properties
        if file_other_meta:
            rm.other_meta.extend(file_other_meta)
        if not filename:
            filename = mk_fn(s=title, dt=dt, root=root)
        else:
            filename = os.path.join(root, filename)
        if os.path.isfile(filename) and not overwrite:
            raise Exception(f"There is already a file named {filename}!")
        node.dump_root(filename)
        logging.info("Created file %s for node %s", filename, title)

    ts = time.time()
    return Zettel(
        uuid=zid,
        title=title,
        filename=filename,
        updated_ts=ts,
        lastchecked_ts=ts,
        zettel_hash=node.node_hash(),
        level=node.actual_level(),
        lineno=1,
        tags=node.tags,
        aliases=aliases if aliases else set(),
        olp=node.olp(),
        properties=node.properties,
    )


def mk_fn(s: str, dt: datetime.datetime, root: str) -> str:
    slug = slugify(s, max_length=50)
    s0 = f"{dt.strftime('%Y%m%d%H%M%S')}-{slug}"
    fn = safe_filename_no_diacritic(f"{s0}.org")
    i = 0
    # Make sure the file doesn't exist
    while True:
        ffn = os.path.join(root, fn)
        if not os.path.isfile(ffn):
            return ffn
        i += 1
        fn = safe_filename_no_diacritic(f"{s0}-{i}.org")


def safe_filename(fn: str, rchar="_") -> str:
    fn = "".join(rchar if c in FORBIDDEN_CHARS else c for c in fn)
    return fn


def safe_filename_no_diacritic(fn: str) -> str:
    return safe_filename(remove_diacritics(fn.strip()))


def remove_diacritics(s: str) -> str:
    return "".join(
        c
        for c in unicodedata.normalize("NFKD", s)
        if unicodedata.category(c) != "Mn"
    )



from orgee.orgnode import OrgNode
from orgee.util import clean_text, escape_url

from .zettel import Zettel


def make_zettel_org_heading(
    zettel: Zettel, use_id: bool = True, add_info_func=None, add_file_url=False
) -> OrgNode:
    node = OrgNode()
    if use_id:
        url = f"id:{zettel.uuid}"
    else:
        url = escape_url("file:%s::%d" % (zettel.filename, zettel.lineno))
    title = "[[%s][%s]]" % (url, clean_text(zettel.title))
    if aliases := zettel.aliases:
        # title += " | %s" % " | ".join(aliases)
        title += " | %s" % " | ".join(
            f"[[{url}][{alias}]]" for alias in aliases
        )
    if len(zettel.olp) > 1:
        title = " > ".join(map(clean_text, zettel.olp[:-1])) + " > " + title
    if add_info_func:
        s = add_info_func(zettel)
        if s:
            title += " " + s
    if add_file_url:
        furl = escape_url("file:%s::%d" % (zettel.filename, zettel.lineno))
        title = f"([[{furl}][🔗]]) " + title
    node.title = title
    node.tags = zettel.all_tags
    return node



from __future__ import annotations  # PEP 585

import logging

from .zettelkasten import Zettelkasten


def rename_tag(
    zk: Zettelkasten, old_tag: str, new_tag: str, dry_run=True
) -> int:
    i = 0
    for zettel in zk.zettels:
        if old_tag in zettel.tags:
            if not dry_run:
                node = zettel.orgnode()
                node.tags.remove(old_tag)
                if new_tag:
                    node.tags.add(new_tag)
                node.dump_root(zettel.filename)
            logging.info(
                "Renamed «%s» to «%s» in %s", old_tag, new_tag, zettel.olp_str()
            )
            i += 1
    logging.info(
        "Renamed «%s» to «%s» in %d zettel%s",
        old_tag,
        new_tag,
        i,
        "s" if i > 1 else "",
    )
    return i
